package org.jbpm;

import static org.jbpm.test.JBPMHelper2.loadStatefulKnowledgeSession;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import javax.naming.InitialContext;
import javax.transaction.UserTransaction;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.drools.KnowledgeBase;
import org.drools.builder.KnowledgeBuilder;
import org.drools.builder.KnowledgeBuilderError;
import org.drools.builder.KnowledgeBuilderFactory;
import org.drools.builder.ResourceType;
import org.drools.io.ResourceFactory;
import org.drools.runtime.StatefulKnowledgeSession;
import org.drools.runtime.process.ProcessInstance;
import org.drools.runtime.process.WorkflowProcessInstance;
import org.jbpm.task.TaskService;
import org.jbpm.task.query.TaskSummary;
import org.jbpm.test.JbpmJUnitTestCase2;
import org.junit.Test;

/**
 * This is a sample file to test a process.
 */
public class WorklfowTest extends JbpmJUnitTestCase2
{
	private static Log testLogger = LogFactory.getLog(WorklfowTest.class);

	public WorklfowTest()
	{
		super(true);
		setPersistence(true);
	}
	
	@Test
	public void testProcess() throws Exception
	{
		StatefulKnowledgeSession ksession = createKnowledgeSession("masterworkflow.bpmn");
		int sessionId = ksession.getId();
		TaskService taskService = getTaskService(ksession);
		
		Map<String, Object> variableMap = new HashMap<String, Object>();
		variableMap.put("ruleId", new Integer(1));
		variableMap.put("ruleState", "DRAFT");
		ProcessInstance processInstance = ksession.startProcess("workflow", variableMap);	
		long processInstanceId = processInstance.getId();
		System.out.println("Process instance id is " + processInstance.getId());
		assertProcessInstanceActive(processInstance.getId(), ksession);
		assertNodeTriggered(processInstance.getId(), "Start Process");

		// saturn app execute start task
		List<TaskSummary> list = taskService.getTasksAssignedAsPotentialOwner("appUser", "en-UK");
		TaskSummary task = findTask(list, "Start Workflow", processInstanceId);
		taskService.start(task.getId(), "appUser");
		taskService.complete(task.getId(), "appUser", null);
		assertNodeTriggered(processInstance.getId(), task.getName());
		
		// submitter execute submit task #1
		list = taskService.getTasksAssignedAsPotentialOwner("submitter", "en-UK");
		task = findTask(list, "Submit for Approval", processInstanceId);
		taskService.start(task.getId(), "submitter");
		taskService.complete(task.getId(), "submitter", null);
		assertNodeTriggered(processInstance.getId(), task.getName());

//		UserTransaction ut = (UserTransaction) new InitialContext().lookup("java:comp/UserTransaction");
//		ut.begin();
		
		// approver reject the rule
		list = taskService.getTasksAssignedAsPotentialOwner("approver", "en-UK");
		task = findTask(list, "Reject Rule", processInstanceId);
		taskService.start(task.getId(), "approver");
		taskService.complete(task.getId(), "approver", null);
		assertNodeTriggered(processInstance.getId(), task.getName());
		
		
//		// RESTART THE SESSION
//		ksession.dispose();
//		ksession = getKnowledgeSession(sessionId);
		
//		ut.commit();
		
		// simulating a system restart
		ksession = restoreSession(ksession, true);
		taskService = getTaskService(ksession);

//		ut = (UserTransaction) new InitialContext().lookup("java:comp/UserTransaction");
//		ut.begin();
		
		// submitter execute submit task #2
		list = taskService.getTasksAssignedAsPotentialOwner("submitter", "en-UK");
		task = findTask(list, "Submit for Approval", processInstanceId);
		taskService.start(task.getId(), "submitter");
		taskService.complete(task.getId(), "submitter", null);
		assertNodeTriggered(processInstance.getId(), task.getName());

		// submitter modifies rule
		list = taskService.getTasksAssignedAsPotentialOwner("submitter", "en-UK");
		task = findTask(list, "Modify Rule", processInstanceId);
		taskService.start(task.getId(), "submitter");
		taskService.complete(task.getId(), "submitter", null);
		assertNodeTriggered(processInstance.getId(), task.getName());
		
		// submitter execute submit task #3
		list = taskService.getTasksAssignedAsPotentialOwner("submitter", "en-UK");
		task = findTask(list, "Submit for Approval", processInstanceId);
		taskService.start(task.getId(), "submitter");
		taskService.complete(task.getId(), "submitter", null);
		assertNodeTriggered(processInstance.getId(), task.getName());
		
		// approver approve the rule
		list = taskService.getTasksAssignedAsPotentialOwner("approver", "en-UK");
		task = findTask(list, "Approve Rule", processInstanceId);
		taskService.start(task.getId(), "approver");
		taskService.complete(task.getId(), "approver", null);
		assertNodeTriggered(processInstance.getId(), task.getName());
		
//		ut.commit();
		
		
		assertNodeTriggered(processInstance.getId(), "End");
		System.out.println("Process instance rule state is " + ((WorkflowProcessInstance)processInstance).getVariable("ruleState"));
		assertProcessInstanceCompleted(processInstance.getId(), ksession);
		ksession.dispose();
	}

	@Test
	public void testTransactions() throws Exception
	{
		StatefulKnowledgeSession ksession = createKnowledgeSession("humantask.bpmn");
		TaskService taskService = getTaskService(ksession);
		UserTransaction ut = (UserTransaction) new InitialContext().lookup("java:comp/UserTransaction");
		ut.begin();
		ProcessInstance processInstance = ksession.startProcess("com.sample.bpmn.hello");
		ut.rollback();
		assertNull(ksession.getProcessInstance(processInstance.getId()));
		List<TaskSummary> list = taskService.getTasksAssignedAsPotentialOwner("john", "en-UK");
		assertEquals(0, list.size());
	}
	
	
	protected StatefulKnowledgeSession getKnowledgeSession(int id)
	{
		KnowledgeBase kbase = createKnowledgeBase();
		return loadStatefulKnowledgeSession(kbase, id);
	}
	
	private TaskSummary findTask(List<TaskSummary> taskList, String taskName, long processInstanceId) throws Exception
	{
		for(TaskSummary summary : taskList)
		{
			if(summary.getName().equals(taskName) && processInstanceId == summary.getProcessInstanceId())
			{
				System.out.println("Task " + taskName + " is chosen");
				return summary;
			}
		}
		throw new Exception("Task not found.");
	}
	
	private KnowledgeBase createKnowledgeBase()
	{
		KnowledgeBuilder kbuilder = KnowledgeBuilderFactory.newKnowledgeBuilder();
		kbuilder.add(ResourceFactory.newClassPathResource("saturnmasterworkflow.bpmn"), ResourceType.BPMN2);

		// Check for errors
		if (kbuilder.hasErrors())
		{
			if (kbuilder.getErrors().size() > 0)
			{
				boolean errors = false;
				for (KnowledgeBuilderError error : kbuilder.getErrors())
				{
					testLogger.warn(error.toString());
					errors = true;
				}
				assertFalse("Could not build knowldge base.", errors);
			}
		}
		return kbuilder.newKnowledgeBase();
	}
}